#include <stdio.h>
#include <atomic>
#include <thread>
#include <string>
#include <iostream>

#include <sys/param.h>
#include <sys/cpuset.h>


#include <unistd.h>

#include <boost/interprocess/managed_shared_memory.hpp>
#include <boost/interprocess/sync/named_mutex.hpp>
#include <boost/interprocess/sync/interprocess_sharable_mutex.hpp>
#include <boost/interprocess/sync/named_sharable_mutex.hpp>
#include <boost/unordered_map.hpp>
#include <boost/unordered_set.hpp>
#include <boost/utility.hpp>
#include <boost/interprocess/containers/set.hpp>
#include <boost/interprocess/containers/map.hpp>
#include <boost/interprocess/containers/string.hpp>
#include <boost/interprocess/allocators/private_node_allocator.hpp>
#include <boost/interprocess/allocators/node_allocator.hpp>

namespace bip = boost::interprocess;

namespace shared_memory {

    typedef bip::managed_shared_memory::segment_manager SegmentManager;
    typedef bip::allocator<void, SegmentManager> VoidAllocator;
    typedef bip::node_allocator<void, SegmentManager> VoidNodeAllocator;
    typedef bip::private_node_allocator<void, SegmentManager> PrivateVoidNodeAllocator;
    typedef bip::private_node_allocator<char, SegmentManager> PrivateCharNodeAllocator;
    typedef bip::basic_string<char, std::char_traits<char>, PrivateCharNodeAllocator> String;

    typedef bip::interprocess_mutex MutexType;
    typedef bip::scoped_lock<MutexType> LockType;

    const std::string shmem_seg_name = "SharedMemorySegment";
    extern bip::managed_shared_memory *segment_;
    extern PrivateVoidNodeAllocator *seg_alloc;

    const uint purge_min_recs = 50000;
    const uint purge_recs = 2;
    const uint purge_range = 10;

    long init(long size);

    void remove();


    template<class PayloadType>  // PayloadType could be a set or int or complex struct (with strings)
    class MappedValType {
    public:
        MappedValType() {}

        MappedValType(time_t expires_at, const PrivateVoidNodeAllocator &void_alloc) : payload(void_alloc),
                                                                                       expires_at(expires_at) {}

        MappedValType(const PayloadType &payload, const time_t &expires_at) : payload(payload),
                                                                              expires_at(expires_at) {}

        ~MappedValType() {}

        PayloadType payload;
        time_t expires_at;
    };


    template<class KeyType, class PayloadType>
    class Map {

    public:
        typedef std::pair<const KeyType, MappedValType<PayloadType> > ValueType;
        typedef bip::private_node_allocator<ValueType, SegmentManager> PrivateValueTypeNodeAllocator;
        typedef bip::node_allocator<ValueType, SegmentManager> ValueTypeNodeAllocator;
        typedef bip::allocator<ValueType, SegmentManager> ValueTypeAllocator;
        typedef bip::map<KeyType, MappedValType<PayloadType>, std::less<KeyType>, ValueTypeAllocator> MapImpl;  // #allocator
        typedef typename MapImpl::iterator MapImplIterator;

        Map() {};

        explicit Map(const std::string &name) : map_name_(name) {
            if (segment_ == nullptr) {
                segment_ = new bip::managed_shared_memory(bip::open_only, shmem_seg_name.c_str());
            }
            assert(segment_ != nullptr);
            map_ = segment_->find_or_construct<MapImpl>(map_name_.data())(
                    std::less<KeyType>(),
                    segment_->get_allocator<MappedValType<PayloadType>>());
            assert(map_ != nullptr);
            mutex_ = segment_->find_or_construct<MutexType>(std::string(name + "mutex").data())();
            assert(mutex_ != nullptr);
        }

        ~Map() {
        }

        MutexType &mutex() {
            return *mutex_;
        }

        bool set(const KeyType &k, const PayloadType &val, bool create_only = true, int expires = 0) {
            time_t unix_now = time(NULL), expires_at = 0;
            if (expires) {
                expires_at = unix_now + expires;
            }
            auto it = map_->find(k);
            if (it == map_->end()) {
                auto ins_res = map_->insert(ValueType(k, MappedValType<PayloadType>(val, expires_at)));
                it = ins_res.first;
            } else {
                if ((*it).second.expires_at && ((*it).second.expires_at < unix_now)) { // expired
                    (*it).second.payload = val;
                    (*it).second.expires_at = expires_at;
                } else {
                    if (create_only)
                        return false;
                    (*it).second.payload = val;
                }
            }
            purge_expired(boost::next(it), purge_recs);
            return true;
        }

    protected:
        MapImpl *map_;
        std::string map_name_;
        MutexType *mutex_;

        void purge_expired(MapImplIterator it, const uint count) {
            // tries to purge expired entries
            // called on each successfull "set" and "add" with a current it + 1
            // purge_expired(map_.begin(), map_.size()) will purge all expired records

            if (map_->size() < purge_min_recs) {
                return;
            }

            if (it == map_->end()) {
                return;
            }
            time_t unix_now = time(NULL);
            MapImplIterator cur_it = it;
            uint del = 0;
            for (int i = 0; (i < purge_range) && (del < count) && (cur_it != map_->end()); ++i) {
                if ((*cur_it).second.expires_at < unix_now) {
                    cur_it = map_->erase(cur_it);
                    ++del;
                } else {
                    cur_it++;
                }
            }
            return;
        }
    };

    template<class KeyType, class ValType>
    class MapSet : public Map<KeyType, bip::set<ValType, std::less<ValType>, bip::private_node_allocator<ValType, SegmentManager>>> {

        typedef bip::set<ValType, std::less<ValType>, bip::private_node_allocator<ValType, SegmentManager>> PayloadType;

        using Map<KeyType, PayloadType>::mutex_;
        using Map<KeyType, PayloadType>::map_;
        using ValueType = typename Map<KeyType, PayloadType>::ValueType;
        using Map<KeyType, PayloadType>::purge_expired;

    public:
        MapSet() : Map<KeyType, PayloadType>() {};

        explicit MapSet(const std::string &name) : Map<KeyType, PayloadType>(name) {};

        ~MapSet() {};

        bool add(const KeyType &k, const ValType &val, int expires = 0) {
            time_t unix_now = time(NULL), expires_at = 0;
            if (expires) {
                expires_at = unix_now + expires;
            }
            auto it = map_->find(k);
            if (it == map_->end()) {
                MappedValType<PayloadType> mapped_val(expires_at,
                                                      PrivateVoidNodeAllocator(segment_->get_segment_manager()));
                mapped_val.payload.insert(val);
                auto ins_res = map_->insert(ValueType(k, mapped_val));
                it = ins_res.first;
            } else {
                if ((*it).second.expires_at && ((*it).second.expires_at < unix_now)) { // expired
                    (*it).second.payload.clear();
                    (*it).second.payload.insert(val);
                    (*it).second.expires_at = expires_at;
                } else {
                    (*it).second.payload.insert(val);
                }
            }
            purge_expired(it, purge_recs);
            return true;
        }
    };

    bip::managed_shared_memory *segment_ = nullptr;
    PrivateVoidNodeAllocator *seg_alloc = nullptr;

    long init(long size) {
        if (segment_ == nullptr) {
            segment_ = new bip::managed_shared_memory(bip::open_or_create, shmem_seg_name.c_str(), size);
            assert (segment_ != nullptr);
            seg_alloc = new PrivateVoidNodeAllocator(segment_->get_segment_manager());
            assert (seg_alloc != nullptr);
        }
        return segment_->get_size();
    }

    void remove() {
        bip::shared_memory_object::remove(shmem_seg_name.c_str());
        segment_ = nullptr;
        return;
    }
}

namespace shmem = shared_memory;

using namespace std;

typedef uint64_t KeyType;
typedef uint64_t SetValType;

struct FooStruct {
    FooStruct() : str_a(*shmem::seg_alloc),
                  str_b(*shmem::seg_alloc) {}

    shmem::String str_a;
    shmem::String str_b;
};

int TTL = 5;
int WORKERS = 3;

void upd1(shmem::Map<KeyType, FooStruct> &map_foo,
          shmem::MapSet<KeyType, SetValType> &map_of_sets,
          FooStruct &fstruct) {
    {
        shmem::LockType lock(map_foo.mutex());
        map_foo.set(rand(), fstruct, true, TTL);
    }

    {
        shmem::LockType lock(map_of_sets.mutex());
        map_of_sets.add(rand(), rand(), TTL);
    }
}

int main(int argc, char *argv[]) {
    srand(time(NULL));
    const unsigned long est_shmem_size = 1 * 1500000000 + (rand() % (400 * 1000000)) + 200 * 1000000;

    shmem::remove();

    if (shmem::init(est_shmem_size) != est_shmem_size) {
        return 1;
    }

    shmem::Map <KeyType, FooStruct> map_foo("MapFoo");
    shmem::MapSet <KeyType, SetValType> map_of_sets("MapOfSets");

    unsigned int wrk = WORKERS;

    // parent process
    cpuset_t cpu_set;
    CPU_ZERO(&cpu_set);
    CPU_SET(wrk, &cpu_set);
    cpuset_setaffinity(CPU_LEVEL_WHICH, CPU_WHICH_PID, -1, sizeof(cpu_set), &cpu_set);
    std::cout << "pid: " << getpid() << " worker: " << wrk << " created" << std::endl;
    srand(wrk);

    // child processes
    while (--wrk) {
        pid_t pid = fork();
        if (pid == 0) {
            cpuset_t cpu_set;
            CPU_ZERO(&cpu_set);
            CPU_SET(wrk, &cpu_set);
            cpuset_setaffinity(CPU_LEVEL_WHICH, CPU_WHICH_PID, -1, sizeof(cpu_set), &cpu_set);
            std::cout << "pid: " << getpid() << " worker: " << wrk << " created" << std::endl;
            srand(wrk);
            break;
        }
    }

    // main loop
    for (int i = 0; i < 1000000; ++i) {
        // this is crucial for crash - triggers PrivateVoidNodeAllocator allocs for long strings
        FooStruct fstruct;
        fstruct.str_a = "aaaaaaaaaaaaaaaaaaaaaaaabbbbbbbbbbbbbbbbbbbbbccccccccccccccccccccccccccccc";
        fstruct.str_b = "aaaaaaaaaaaaaaaaaaaaaaaabbbbbbbbbbbbbbbbbbbbbccccccccccccccccccccccccccccd";

        upd1(map_foo, map_of_sets, fstruct);

        if (i % 10000 == 0) {
            std::cout << time(NULL) << " processed " << i << endl;
        }
    }
}
